/*
 * Copyright (c) 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden). 
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. All advertising materials mentioning features or use of this software 
 *    must display the following acknowledgement: 
 *      This product includes software developed by Kungliga Tekniska 
 *      Hgskolan and its contributors. 
 *
 * 4. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
RCSID("$Id: putarg.c,v 1.1 1998/04/07 01:21:40 lha Exp $");
#endif

#include <stdio.h>
#include <roken.h>
#include "putarg.h"

#define ISFLAG(X) ((X).type == parg_flag || (X).type == parg_negative_flag)

void
putarg_usage(struct putargs *args, 
	     size_t num_args, 
	     const char *extra_string)
{
    size_t i ;

    if (extra_string)
	printf("Usage: %s ", extra_string);

    for (i = 0 ; i < num_args ; i++) {
	printf("%s-%s",
	       args[i].mandatoryp?"":"[",
	       args[i].long_name);
	if (args[i].help)
	    printf(" <%s>",
	       args[i].help);
	printf("%s ", args[i].mandatoryp?"":"]");
    }
    printf("\n");
}



static void
add_string(putarg_strings *s, char *value)
{
    s->strings = realloc(s->strings, (s->num_strings + 1) * sizeof(*s->strings));
    s->strings[s->num_strings] = value;
    s->num_strings++;
}

static int
arg_match_long(struct putargs *args, size_t num_args,
	       char **argv)
{
    int i;
    char *optarg = NULL;
    int negate = 0;
    int partial_match = 0;
    struct putargs *partial = NULL;
    struct putargs *current = NULL;
    int argv_len;
    char *p;

    argv_len = strlen(*argv);
    p = strchr (*argv, '=');
    if (p != NULL)
	argv_len = p - *argv;

    for (i = 0; i < num_args; ++i) {
	if(args[i].long_name) {
	    int len = strlen(args[i].long_name);
	    char *p = *argv + 1;
	    int p_len = argv_len;
	    negate = 0;

	    for (;;) {
		if (strncmp (args[i].long_name, p, len) == 0) {
		    current = &args[i];
		    optarg  = *(argv + 1);
		} else if (strncmp (args[i].long_name,
				    p,
				    p_len) == 0) {
		    ++partial_match;
		    partial = &args[i];
		    optarg  = *(argv + 1);
		} else if (ISFLAG(args[i]) && strncmp (p, "no-", 3) == 0) {
		    negate = !negate;
		    p += 3;
		    p_len -= 3;
		    continue;
		}
		break;
	    }
	    if (current)
		break;
	}
    }
    if (current == NULL) {
	if (partial_match == 1)
	    current = partial;
	else
	    return PARG_ERR_NO_MATCH;
    }

    if(*optarg == '\0' && !ISFLAG(*current))
	return PARG_ERR_NO_MATCH;

    return parg_assign(current, optarg);
}


int
parg_assign(struct putargs *args, char *arg)
{
    if(args->type == parg_flag){
	*(int*)args->value = 1;
	return 0;
    }
    if(args->type == parg_negative_flag){
	*(int*)args->value = 0;
	return 0;
    }
    if(arg == NULL)
	return PARG_ERR_NO_ARG;
    if(args->type == parg_integer){
	int tmp;
	if(sscanf(arg, "%d", &tmp) != 1)
	    return PARG_ERR_BAD_ARG;
	*(int*)args->value = tmp;
	goto out;
    }else if(args->type == parg_string){
	*(char**)args->value = arg;
			    goto out;
    }else if(args->type == parg_strings){
	add_string((putarg_strings*)args->value, arg);
	goto out;
    }
    return PARG_ERR_BAD_ARG;
    
out:
    return 0;
}



int
putarg(struct putargs *args, size_t num_args, 
       int argc, char **argv, int *optind)
{
    int i, swcount = *optind ;
    int ret = 0;

    srand (time(NULL));
    (*optind)++;
    for(i = *optind; (num_args > swcount|| swcount == -1) && i < argc; i++) {

	if(argv[i][0] != '-' && swcount != -1) {
	    ret = parg_assign(&args[swcount], argv[i]);
	    swcount++;
	    if (ret)
		break;
	}else if (argv[i][0] == '-' && argc > i+1) {
	    swcount = -1 ;
	    ret = arg_match_long (args, num_args, &argv[i]);
	    if(ret)
		break;
	    i++;
	} else {
	    ret = PARG_ERR_BAD_ARG;
	    break;
	}
    }
    *optind = i;
    return ret;
}

#if TEST
int foo_flag = 2;
int flag1 = 0;
int flag2 = 0;
int bar_int;
char *baz_string;

struct putargs args[] = {
    { NULL,  parg_flag, &flag1, "one", NULL },
    { NULL,  parg_flag, &flag2, "two", NULL },
    { "foo", parg_negative_flag, &foo_flag, "foo", NULL },
    { "bar", parg_integer, &bar_int, "bar", "seconds"},
    { "baz", parg_string, &baz_string, "baz", "name" },
};

int main(int argc, char **argv)
{
    int optind = 0;
    while(putarg(args, 5, argc, argv, &optind))
	printf("Bad arg: %s\n", argv[optind]);
    printf("flag1 = %d\n", flag1);  
    printf("flag2 = %d\n", flag2);  
    printf("foo_flag = %d\n", foo_flag);  
    printf("bar_int = %d\n", bar_int);
    printf("baz_flag = %s\n", baz_string);
/*    arg_printusage (args, 5, "nothing here"); */
    return 0;
}
#endif
